/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/credentials/CredentialPersistenceManager.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc. 
 * All rights reserved.
 */
package com.netapp.common.credentials;

import java.util.ArrayList;
import java.util.List;

import org.apache.commons.codec.DecoderException;
import org.apache.commons.codec.binary.Hex;

import com.netapp.common.MsgKey;
import com.netapp.common.encryption.EncryptionException;
import com.netapp.common.encryption.SymmetricEncryption;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.StringUtils;

/**
 * Manages access to the credential store. For external views, the password
 * is in plain text. Internally, passwords are encrypted using a symmetric key.
 * 
 * Each time a new credential is to be persisted, the credentials must be validated
 * through the implementation of a CredentialValidator. These validators are 
 * implemented and configured by the application.
 * 
 * The following configuration properties must be set
 * 
 * <ul>
 *   <li>credential.symmetric.password
 *   <li>credential.symmetric.algorithm (defaults to AES)
 * </ul>
 * 
 * The following configuraiton properties are optional
 * 
 * <ul>
 *   <li>credential.validator.# (1-n, sequentially)
 * </ul>
 * 
 * @author joshb
 *
 */
public abstract class CredentialPersistenceManager {
    private static final NALogger LOGGER = NALogger.getLogger(CredentialPersistenceManager.class);

    private List<CredentialValidator> validators;
    private ConfigBundle context;
    private String symmetricAlgorithm;
    private boolean caseInsensitive;

    public CredentialPersistenceManager() {
    }
    
    /** 
     * Init method. To be called after all configuration is complete, generally
     * by Spring or another framework.
     *
     */
    public void init() throws CredentialPersistenceException {
        loadConfiguration();
        loadValidators(getContext());
    }

    /**
     * Check if there is a validator for the specified type. If so, run it.
     * If the credential validates, then persist it.
     * 
     * @param credential
     * @param type
     * @throws CredentialPersistenceException
     */
    public void addCredential(Credential credential, String type) throws CredentialPersistenceException {
        if (null == credential || null == credential.getKey() || StringUtils.isEmpty(type)) {
            throw new CredentialPersistenceException(LOGGER
                .warn(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_ERROR_INVALID_INPUTS));
        }

        for (CredentialValidator validator : getValidators()) {
            if (validator.getType().equalsIgnoreCase(type)) {
                if (!validator.isValid(credential)) {
                    throw new CredentialPersistenceException(LOGGER.warn(
                        MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_ERROR_VALIDATION_FAILED, credential, type, validator
                            .getClass().getName()));
                }
            }
        }

        persistCredentialToStore(credential, type);
    }

    /**
     * Retrieve the credential for the specified key and type. If the credential does not
     * exist, return null.
     * 
     * @param key
     * @param type
     * @throws CredentialPersistenceException
     * @return Credential for the requested <code>{key, type}</code> tuple, or null if no such credential exists.
     */
    public Credential retrieveCredential(CredentialKey key, String type) throws CredentialPersistenceException {
        validateKeyAndType(key, type);
        
        String password = retreivePasswordFromStore(key, type);
        
        return new Credential(key, password);
    }

    /**
     * Attempt to remove the credential for the specified key and type. If the credential
     * cannot be removed, return false.
     * 
     * @param key
     * @param type
     * @throws CredentialPersistenceException
     */
    public void removeCredential(CredentialKey key, String type) throws CredentialPersistenceException {
        validateKeyAndType(key, type);

        if (null != retreivePasswordFromStore(key, type)) {
            removeCredentialFromStore(key, type);
        }
    }
    
    /**
     * List all of the credentials found within the credential persistence manager. If the type is specified, only
     * list credentials of that particular type. If the type is null, list all credentials.
     * 
     * @param type
     * @return List of credentials of <code>type</code>, or all known credentials if <code>type</code> is null.
     * @throws CredentialPersistenceException
     */
    public List<Credential> listCredentials(String type) throws CredentialPersistenceException {
        return listCredentialsFromStore(type);
    }

    /**
     * The contract is such that the password must be encrypted before persisting.
     * 
     * @param credential
     * @param type
     * @throws CredentialPersistenceException
     */
    protected abstract void persistCredentialToStore(Credential credential, String type)
        throws CredentialPersistenceException;

    /**
     * If the credential information exists in the persistent store, remove the entry.
     * 
     * @param key
     * @param type
     * @throws CredentialPersistenceException
     */
    protected abstract void removeCredentialFromStore(CredentialKey key, String type)
        throws CredentialPersistenceException;

    /**
     * Retrieve the unencrypted password from the persistent store. If the credential/type 
     * does not exist in the persistent store, return null.
     * 
     * @param key
     * @param type
     * @return Unencrypted password for the given <code>{key, type}</code> tuple, or null if no such record exists
     * @throws CredentialPersistenceException
     */
    protected abstract String retreivePasswordFromStore(CredentialKey key, String type)
        throws CredentialPersistenceException;
    
    /**
     * List all of the credentials from the persistent store. If the type is null, list all
     * credentials, otherwise filter the credentials for only the specified type.
     * 
     * @param type
     * @return List of credentials of <code>type</code>, or all known credentials if <code>type</code> is null.
     * @throws CredentialPersistenceException
     */
    protected abstract List<Credential> listCredentialsFromStore(String type)
        throws CredentialPersistenceException;

    /**
     * Given an original password, use a symmetric encryption algorithm to encrypt the
     * password. The secret key and algorithm can be configured in the CredentialPersistenceManager.
     * The output is the encrypted value represented as hex.
     * 
     * @param original
     * @return Encrypted password
     * @throws CredentialPersistenceException
     */
    protected String encrypt(String original) throws CredentialPersistenceException {
        if (null == original) {
            return null;
        }
        
        try {
            String passwordHex = CredentialConstants.ENCRYPTION_PASSWORD.lookup(getContext());
            byte[] secret = Hex.decodeHex(passwordHex.toCharArray());

            // call the cipher using our cipher password and algorithm to encrypt the original string
            byte[] encrypted = SymmetricEncryption.encrypt(getSymmetricAlgorithm(), secret, original);

            // encode the outgoing string in hex
            return String.valueOf(Hex.encodeHex(encrypted));
        }
        catch (EncryptionException e) {
            throw new CredentialPersistenceException(e);
        }
        catch (DecoderException e) {
            throw new CredentialPersistenceException(e);
        }
    }

    /**
     * Given an encrypted password stored in hex values, use a symmetric encryption algorithm to decrypt the password. 
     * The secret key and algorithm can be configured in the CredentialPersistenceManager. The 
     * output is the original password.
     * 
     * @param encrypted
     * @return Plaintext of encrypted
     * @throws CredentialPersistenceException
     */
    protected String decrypt(String encrypted) throws CredentialPersistenceException {
        if (null == encrypted) {
            return null;
        }
        
        try {
            String passwordHex = CredentialConstants.ENCRYPTION_PASSWORD.lookup(getContext());
            byte[] secret = Hex.decodeHex(passwordHex.toCharArray());

            // decode the incoming string from hex
            byte[] encryptedBytes = Hex.decodeHex(encrypted.toCharArray());

            // call the cipher using our cipher password and algorithm to decrypt the original string
            return SymmetricEncryption.decrypt(getSymmetricAlgorithm(), secret, encryptedBytes);
        }
        catch (EncryptionException e) {
            throw new CredentialPersistenceException(e);
        }
        catch (DecoderException e) {
            throw new CredentialPersistenceException(e);
        }
    }

    /**
     * Check and set the configuration variables.
     * 
     * @throws CredentialPersistenceException
     */
    protected void loadConfiguration() throws CredentialPersistenceException {
        String passwordHex = CredentialConstants.ENCRYPTION_PASSWORD.lookup(getContext());
        String algorithm = CredentialConstants.ENCRYPTON_ALGORITHM.lookup(getContext());

        if (null == passwordHex || StringUtils.isEmpty(passwordHex)) {
            throw new CredentialPersistenceException(LOGGER
                .error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_CONFIGURATION_ERROR_SECRET_KEY_NOT_DEFINED));
        }

        if (null == algorithm || StringUtils.isEmpty(algorithm)) {
            throw new CredentialPersistenceException(LOGGER
                .error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_CONFIGURATION_ERROR_ALGORITHM_NOT_DEFINED));
        }

        setSymmetricAlgorithm(algorithm);
        
        boolean insensitive = CredentialConstants.CASE_INSENSITVIE.lookup(getContext());
        setCaseInsensitive(insensitive);
    }

    /**
     * load the configured validators from the configuration bundle supplied.
     * Validators are defined by the configuration property 'credential.validator.<#>'. 
     * Each validator that is defined with this key is loaded. The numbers start at
     * one (1) and increase sequentially. The first entry that returns null causes the
     * list of validators to stop loading.
     *  
     * @param config
     */
    protected void loadValidators(ConfigBundle config) throws CredentialPersistenceException {
        int index = 1;
        String className = null;

        while (null != (className = config.getProperty(CredentialConstants.CONFIG_KEY_PREFIX_VALIDATORS + index))) {
            String property = CredentialConstants.CONFIG_KEY_PREFIX_VALIDATORS + index;
            index++;

            try {
                Class<?> validator = Class.forName(className);

                if (!ClassUtils.typeAssignableFrom(CredentialValidator.class, validator)) {
                    LOGGER.error(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_CONFIGURATION_ERROR_INVALID_VALIDATOR_CLASS,
                        property, className, CredentialValidator.class.getName());
                    continue;
                }

                getValidators().add((CredentialValidator) validator.newInstance());
            }
            catch (ClassNotFoundException e) {
                throw new CredentialPersistenceException(e);
            }
            catch (InstantiationException e) {
                throw new CredentialPersistenceException(e);
            }
            catch (IllegalAccessException e) {
                throw new CredentialPersistenceException(e);
            }
        }

        LOGGER.debug(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_CONFIGURATION_INFO_LOADED_VALIDATORS, getValidators().size());
    }

    protected void validateKeyAndType(CredentialKey key, String type) throws CredentialPersistenceException {
        if (null == key || StringUtils.isEmpty(type)) {
            throw new CredentialPersistenceException(LOGGER.warn(MsgKey.CREDENTIAL_PERSISTENCE_MANAGER_ERROR_INVALID_INPUTS));
        }
    }
    
    protected List<CredentialValidator> getValidators() {
        if (null == validators) {
            validators = new ArrayList<CredentialValidator>();
        }

        return validators;
    }

    protected void setValidators(List<CredentialValidator> validators) {
        this.validators = validators;
    }

    protected ConfigBundle getContext() {
        return context;
    }

    public void setContext(ConfigBundle context) {
        this.context = context;
    }

    protected String getSymmetricAlgorithm() {
        return symmetricAlgorithm;
    }

    protected void setSymmetricAlgorithm(String symmetricAlgorithm) {
        this.symmetricAlgorithm = symmetricAlgorithm;
    }

    protected boolean isCaseInsensitive() {
        return caseInsensitive;
    }

    protected void setCaseInsensitive(boolean caseInsensitive) {
        this.caseInsensitive = caseInsensitive;
    }

}
