/*
 * $Id$
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.execution;

import java.util.ArrayList;

import com.netapp.common.util.CollectionUtils;
import com.netapp.common.util.StringUtils;
import com.netapp.common.util.Transformer;

/**
 * {@link Transformer} for transforming {@link ShellRunnable}'s.  Note that ShellCommandTransformer's
 * can be chained together.
 * 
 * @author claird
 * @version $Revision: #2 $ $Date: Jul 2, 2008 $
 */
abstract class ShellCommandTransformer implements Transformer<ShellRunnable, ShellRunnable> {
    
    private ShellCommandTransformer shellCommandTransformer;

    /*
     * List of special characters that should not be quoted
     * It is assumed that if a ShellCommand is created using these characters that the caller
     * intends them to be used as special shell characters.
     * 
     * This means that files that include these characters will unfortunately not get handled correctly
     * 
     * Ideally we would refactor ShellCommand to make such character usages explicit in the creation
     * of the command so that this intent is passed to the QUOTING_TRANSFORMER.  But that's for another time.
     */
    private static final String[] SPECIAL_SHELL_CHARACTERS = new String[] { "?", "\"", "|", "`", "$", ";", "&", "*" };

    /**
     * Quotes any string that does not contain SPECIAL_SHELL_CHARACTERS
     */
    protected static final Transformer<String, String> QUOTING_TRANSFORMER = new Transformer<String, String>() {
        public String transform(String s) {
            if (StringUtils.containsAny(s, SPECIAL_SHELL_CHARACTERS)) {
                return s;
            }
            return "\"" + s + "\"";
        }
    };

    public ShellCommandTransformer() {
        super();
    }

    public ShellCommandTransformer(ShellCommandTransformer shellCommandTransformer) {
        this();
        this.shellCommandTransformer = shellCommandTransformer;
    }

    /**
     * @see com.netapp.common.util.Transformer#transform(java.lang.Object)
     */
    public ShellRunnable transform(ShellRunnable shellCommand) {
        ShellRunnable shellCommandToTransform = shellCommand;
        if (this.shellCommandTransformer != null) {
            shellCommandToTransform = this.shellCommandTransformer.transform(shellCommand);
        }

        return this.performTransformation(shellCommandToTransform);
    }

    protected String createQuotedCommandString(ShellRunnable shellCommand) {
        return CollectionUtils.collectionToString(CollectionUtils.transform(shellCommand.getArgumentList(),
            new ArrayList<String>(), QUOTING_TRANSFORMER), " ");
    }

    protected abstract ShellRunnable performTransformation(ShellRunnable shellCommand);

}
