/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/flow/Engine.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.flow;

import static com.netapp.common.flow.FlowConstants.DEFAULT_THREADPOOL_IDLE;
import static com.netapp.common.flow.FlowConstants.DEFAULT_THREADPOOL_SIZE;
import static com.netapp.common.flow.FlowConstants.DIAGNOSTIC_AUTO_ENABLE;
import static com.netapp.common.flow.FlowConstants.DIAGNOSTIC_OPERATION_NAME;
import static com.netapp.common.flow.FlowConstants.FAILED_OPERATION_CLIENT_REQUEST_ID;
import static com.netapp.common.flow.FlowConstants.FAILED_OPERATION_ID;
import static com.netapp.common.flow.FlowConstants.FAILED_OPERATION_REGISTERS;
import static com.netapp.common.flow.FlowConstants.FAILED_OPERATION_TYPE;
import static com.netapp.common.util.ConfigProperty.intProperty;

import java.lang.management.ManagementFactory;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.management.InstanceAlreadyExistsException;
import javax.management.MBeanRegistrationException;
import javax.management.MBeanServer;
import javax.management.MalformedObjectNameException;
import javax.management.NotCompliantMBeanException;
import javax.management.ObjectName;

import com.netapp.common.MsgKey;
import com.netapp.common.flow.api.OperationState;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ConfigBundle;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.ListenerProxy;

/**
 * Main external entry point to the workflow engine.
 * It is generally expected this engine will be used over some sort of RPC
 * connection like SOAP that ensures object references supplied as inputs or
 * outputs are decoupled from the caller by serialization. In addition, the external
 * visibility of references used inside a workflow may depend on the PersistenceManager used.
 * For performance reasons, no explicit effort is made to decouple references passed in.
 * Callers using this subsystem in-process should keep this in mind and design accordingly.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Engine {
    private static NALogger logger = NALogger.getLogger(Engine.class);
    
    private String name;
    private PersistenceManager manager;
    private EngineConfiguration configuration;
    private ConfigBundle applicationContext;
    private Map<String, Threadpool> queues;
    private ListenerProxy<LifecycleListener> operationListeners = new ListenerProxy<LifecycleListener>(LifecycleListener.class);
    private DiagnosticListener diagnosticListener;
    private boolean started;
    
    /**
     * Create a new workflow engine.
     * This method is designed for use via Bean frameworks like Spring.
     * Prior to using any of the operation methods, you must supply a configuration
     * via {@link #setConfiguration(EngineConfiguration)} and {@link #setPersistenceManager(PersistenceManager)}
     * and then calling {@link #start()}.
     */
    public Engine() {
    }
    
    /**
     * Create and start a new workflow engine.
     * @param configuration Engine configuration data
     * @param manager Persistence Manager
     */
    public Engine(EngineConfiguration configuration, PersistenceManager manager) {
        this.configuration = configuration;
        this.manager = manager;
        this.start();
    }
    
    /**
     * Set the configuration for this engine.
     * @param configuration
     */
    public void setConfiguration(EngineConfiguration configuration) {
        assert this.configuration == null;
        this.configuration = configuration;
    }
    
    public EngineConfiguration getConfiguration() {
        return this.configuration;
    }
    
    /**
     * Set the persistence manager for this engine.
     * @param manager
     */
    public void setPersistenceManager(PersistenceManager manager) {
        assert this.manager == null;
        logger.internal("Using PersistenceManager %s", manager.getClass().getSimpleName());
        this.manager = manager;
    }
    
    protected PersistenceManager getPersistenceManager() {
        return this.manager;
    }
    
    /**
     * Set the application context for this engine. The application context is a
     * {@link com.netapp.common.util.ConfigBundle} for storing application-specific config values.
     */
    public void setApplicationContext(ConfigBundle applicationContext) {
        this.applicationContext = applicationContext;
    }
    
    public ConfigBundle getApplicationContext() {
        return this.applicationContext;
    }
    
    public void setName(String name) {
        assert !this.started;
        this.name = name;
    }
    
    /**
     * Stop this engine.
     * {@link #start()} must be called prior to executing this method.
     */
    public synchronized void stop() {
        if (!this.started) {
            return;
        }
        MBeanServer mbeanServer;
        mbeanServer = ManagementFactory.getPlatformMBeanServer();
        
        for (Threadpool pool : this.queues.values()) {
            pool.shutdown();
            
            try {
                ObjectName beanName = new ObjectName("com.netapp.common.flow:"
                    + ((this.name != null) ? "engine=" + this.name + "," : "")
                    + "type=Threadpool,name=" + pool.getName());
                
                mbeanServer.unregisterMBean(beanName);
            }
            // CHECKSTYLE:OFF
            catch (Exception e) {
            // CHECKSTYLE:ON
                logger.warn(MsgKey.FLOW_JMX_STOP_EXCEPTION, e);
            }
        }
        this.manager.stop();
    }
    
    /**
     * Start this engine.
     * {@link #setConfiguration(EngineConfiguration)} and {@link #setPersistenceManager(PersistenceManager)}
     * must be called prior to executing this method.
     *
     */
    public synchronized void start() {
        if (this.started) {
            return;
        }
        assert this.configuration != null;
        assert this.manager != null;
        this.manager.start(this);
        if (!this.manager.isValid()) {
            // We should throw an exception here and just refuse to start,
            // but Jetty won't bubble it up to us. See burt 256580
            logger.error(MsgKey.FLOW_START_MANAGER_INVALID);
            return;
        }
        
        this.queues = new HashMap<String, Threadpool>();
        for (OperationTemplate t : this.configuration.getOperationTemplates()) {
            String poolName = t.getThreadpool();
            Threadpool pool = this.queues.get(poolName);
            if (pool == null) {
                int size = DEFAULT_THREADPOOL_SIZE;
                int idle = DEFAULT_THREADPOOL_IDLE;
                if (this.applicationContext != null) {
                    ConfigProperty<Integer> sizeProp = intProperty("flow.threadpool." + poolName + ".size", size);
                    ConfigProperty<Integer> idleProp = intProperty("flow.threadpool." + poolName + ".idle", idle);
                    size = sizeProp.lookup(this.applicationContext);
                    idle = idleProp.lookup(this.applicationContext);
                }
                pool = new Threadpool(poolName, size);
                pool.setIdleTime(idle);
                this.queues.put(poolName, pool);
            }
            pool.addValidOperation(t.getName());
        }
        
        try {
            enableJMX();
        }
        // CHECKSTYLE:OFF
        catch (Exception e) {
        // CHECKSTYLE:ON
            logger.warn(MsgKey.FLOW_JMX_START_EXCEPTION, e);
        }
        
        if (this.applicationContext != null) {
            setAutomaticDiagnostics(DIAGNOSTIC_AUTO_ENABLE.lookup(this.applicationContext));
        }
        logger.internal("Flow Engine configuration complete - Starting queues");
        for (Threadpool pool : queues.values()) {
            pool.start(this);
        }
        this.started = true;
    }
    
    /**
     * Get a list of configured operations this engine has
     * @return Configured operations
     */
    public List<OperationInformation> getRegisteredOperations() {
        assert this.configuration != null;
        return this.configuration.getRegisteredOperations();
    }
    
    /**
     * Assert that the engine is started and the manager is running, or throw a FlowException
     * @throws FlowException The engine is in an invalid state to accept requests
     */
    private void assertValidState() {
        if (!this.started) {
            throw new FlowException(logger.error(MsgKey.FLOW_DENY_ENGINE));
        }
        if (!this.manager.isValid()) {
            throw new FlowException(logger.error(MsgKey.FLOW_DENY_PERSISTENCE_MANAGER));
        }
    }
    
    /**
     * Create and queue a new operation.
     * @param operationName Name of the requested operation
     * @param parameters Map of initial input parameter names and values
     * @return ID of the new operation
     */
    public OperationId operationStart(String operationName, Map<String, ?> parameters) {
        assertValidState();
        
        OperationTemplate operationTemplate = this.configuration.getOperationTemplate(operationName);
        if (operationTemplate == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_OPERATION_OPERATION_NOT_DEFINED, operationName));
        }
        OperationContext context = this.manager.createContext(operationTemplate, parameters);
        
        context.enterContext();
        operationListeners.proxy().operationCreated(context.getId(), context.getClientRequestId());
        Threadpool pool = queues.get(operationTemplate.getThreadpool());
        logger.internal("Waking up queue %s for new operation %s", pool, context);
        pool.notifyScheduler();
        context.exitContext();
        return context.getId();
    }
    
    /**
     * Return the current status and final results for an operation.
     * @param id Operation id to retrieve
     * @return structure containing the final state and outputs of the requested operation
     */
    public Result operationResult(String id) {
        assertValidState();
        
        Result result = null;
        OperationContext context = this.manager.getContext(id);
        if (context != null) {
            result = context.getResult();
        }
        else {
            result = new Result();
            if (OperationId.isId(id)) {
                result.setId(id);
            }
            else {
                result.setClientRequestId(id);
            }
            result.setState(OperationState.INVALID_ID);
        }
        return result;
    }
    
    /**
     * Attempt to cancel an operation. This only requests the operation
     * be canceled at the next cancelable point. The caller should continue to
     * call {@link #operationResult} to determine the final operation state.
     * 
     * @param id Operation to cancel
     */
    public void operationCancel(String id) {
        assertValidState();
        this.manager.cancelContext(id);
    }
    
    public void operationRelease(String id) {
        assertValidState();
        this.manager.releaseContext(id);
    }
    
    public OperationId operationDiagnostic(String id) {
        assertValidState();
        OperationTemplate operationTemplate = this.configuration.getDiagnosticTemplate();
        if (operationTemplate == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_DIAGNOSTIC_OPERATION_NOT_DEFINED));
        }
        
        OperationContext failedContext = this.manager.getContext(id);
        if (failedContext == null) {
            throw new FlowException(logger.error(MsgKey.FLOW_DIAGNOSTIC_OPERATION_ID_NOT_FOUND, id));
        }
        if (!failedContext.isTerminalState()) {
            throw new FlowException(logger.error(MsgKey.FLOW_DIAGNOSTIC_OPERATION_NOT_TERMINAL, id));
        }
        
        logger.internal("Initializing Diagnostic context for " + failedContext.getId());
        Map<String, Object> params = new HashMap<String, Object>();
        params.put(FAILED_OPERATION_ID, failedContext.getId().getId());
        params.put(FAILED_OPERATION_CLIENT_REQUEST_ID, failedContext.getClientRequestId());
        params.put(FAILED_OPERATION_TYPE, failedContext.getOperationType());
        params.put(FAILED_OPERATION_REGISTERS, new HashMap<String, Object>(failedContext.getPersistentRegisters()));
        
        OperationContext context = this.manager.createContext(operationTemplate, params);
        
        context.enterContext();
        Threadpool pool = queues.get(operationTemplate.getThreadpool());
        logger.internal("Waking up queue %s for new operation %s", pool, context);
        pool.notifyScheduler();
        context.exitContext();
        return context.getId();
    }
    
    /**
     * Add a new listener to be notified during Operation lifecycle events
     * Listeners are weakly referenced and will not be called if the listener
     * is garbage collected.
     * @param listener
     */
    public void addLifecycleListener(LifecycleListener listener) {
        this.operationListeners.addListener(listener);
    }
    
    protected LifecycleListener lifecycleListenerProxy() {
        return this.operationListeners.proxy();
    }
    
    /**
     * Enable JMX Management objects for this engine.
     * @throws MalformedObjectNameException
     * @throws InstanceAlreadyExistsException
     * @throws MBeanRegistrationException
     * @throws NotCompliantMBeanException
     */
    private void enableJMX() throws MalformedObjectNameException, InstanceAlreadyExistsException, MBeanRegistrationException, NotCompliantMBeanException {
        logger.debug(MsgKey.FLOW_JMX_STARTING);
        MBeanServer mbeanServer;
        mbeanServer = ManagementFactory.getPlatformMBeanServer();
        
        ObjectName beanName;
        for (Threadpool pool : queues.values()) {
            beanName = new ObjectName("com.netapp.common.flow:"
                + ((this.name != null) ? "engine=" + this.name + "," : "")
                + "type=Threadpool,name=" + pool.getName());
            mbeanServer.registerMBean(pool, beanName);
        }  
    }
    
    public void setAutomaticDiagnostics(boolean automaticDiagnostics) {
        if (automaticDiagnostics 
                && (this.configuration.getDiagnosticTemplate() != null) 
                && (this.diagnosticListener == null)) {
            this.diagnosticListener = new DiagnosticListener();
            this.addLifecycleListener(diagnosticListener);
        }
        else if (!automaticDiagnostics 
                && (this.diagnosticListener != null)) {
            this.diagnosticListener.disable();
            this.diagnosticListener = null;
        }
    }
    
    private class DiagnosticListener extends LifecycleAdapter {
        private boolean enabled = true;
        
        public void operationCompleted(OperationId id, String clientRequestId) {
            OperationContext context = Engine.this.manager.getContext(id.getId());
            if (this.enabled
                    && !OperationState.COMPLETE.equals(context.getState())
                    && !context.getOperationType().equals(DIAGNOSTIC_OPERATION_NAME)) {
                try {
                    OperationId diag = Engine.this.operationDiagnostic(context.getId().getId());
                    logger.warn(MsgKey.FLOW_OPERATION_AUTO_DIAGNOSTIC, diag);
                }
                catch (FlowException e) {
                    logger.internal("Error creating diagnostic for %s: %s", e, context, e.getLocalizedMessage());
                }
            }
        }
        
        private void disable() {
            this.enabled = false;
        }
    }
}
