/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/logging/NALogger.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.logging;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import com.netapp.common.util.ListenerProxy;

/**
 * Log messages through the logging framework.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */

public class NALogger {
    private static Map<String, NALogger> loggers = new ConcurrentHashMap<String, NALogger>();
    private static ListenerProxy<LogListener> listeners = new ListenerProxy<LogListener>(LogListener.class);
    private static LogAdapterFactory factory;
    static {
        // TODO: Add ability to specify a different factory
        // via a system property or file in the classpath
        try {
            factory = new Log4jAdapterFactory();
        }
        // CHECKSTYLE:OFF
        catch (Throwable t) { }
        // CHECKSTYLE:ON
        if (factory == null) {
            factory = new JdkLoggerAdapterFactory();
        }
    }
    
    private LogAdapter adapter;
    
    /** Class that instantiated this logger. */
    String instantiatingClass;
 
    private NALogger(String clazz) {
        this.instantiatingClass = clazz;
        this.adapter = factory.createLogAdapter(instantiatingClass);
    }
 
    /**
     * Get a logger for the supplied class
     * @param clazz
     * @return logger for the supplied class
     */
    public static NALogger getLogger(Class<?> clazz) {
        return getLogger(clazz.getName());
    }
 
    public static NALogger getLogger(String clazz) {
        NALogger logger = loggers.get(clazz);
        if (logger == null) {
            logger = new NALogger(clazz);
            loggers.put(clazz, logger);
        }
        return logger;
    }
    /**
     * Get a logger for the calling context automatically.
     * @return Logger associated to the calling class
     */
    public static NALogger getLogger() {
        Throwable t = new Throwable();
        String caller = t.getStackTrace()[1].getClassName();         
        return getLogger(caller);
    }
 
    /**
     * Add a new listener to be notified when a message is logged.
     * Listeners are weakly referenced and will not be called if the listener
     * is garbage collected.
     * @param listener
     */
    public static void addListener(LogListener listener) {
        listeners.addListener(listener);
    }
    
    protected static void notifyListeners(LogMessage message) {
        listeners.proxy().append(message);
    }
 
    public String internal(String msg) {
        adapter.log(LogLevel.DEBUG, msg);
        return msg;
    }
    
    public String internal(String msg, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.DEBUG, formattedMsg);
        return formattedMsg;
    }

    public String internal(String msg, Throwable t, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.DEBUG, formattedMsg, t);
        return formattedMsg;
    }
    
    public String internalTrace(String msg) {
        adapter.log(LogLevel.TRACE, msg);
        return msg;
    }
    
    public String internalTrace(String msg, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.TRACE, formattedMsg);
        return formattedMsg;
    }

    public String internalTrace(String msg, Throwable t, Object... params) {
        String formattedMsg = String.format(msg, params);
        adapter.log(LogLevel.TRACE, formattedMsg, t);
        return formattedMsg;
    }
    
    String log(LogMessage msg) {
        adapter.log(msg);
        notifyListeners(msg);
        return msg.localizedMessage();
    }
    
    String log(LogMessage msg, Throwable t) {
        adapter.log(msg, t);
        notifyListeners(msg);
        return msg.localizedMessage();
    }
    
    /*
     * standard logging methods
     */
    
    public String log(LogLevel level, MessageKey key) {
        LogMessage msg = new LogMessage(level, this.instantiatingClass, key, (Object[])null);
        return log(msg);
    }
    
    public String log(LogLevel level, MessageKey key, Throwable t) {
        LogMessage msg = new LogMessage(level, this.instantiatingClass, key, (Object[])null);
        return log(msg, t);
    }
    
    public String log(LogLevel level, MessageKey key, Object... params) {
        LogMessage msg = new LogMessage(level, this.instantiatingClass, key, params);
        return log(msg);
    }
    
    public String log(LogLevel level, MessageKey key, Throwable t, Object... params) {
        LogMessage msg = new LogMessage(level, this.instantiatingClass, key, params);
        return log(msg, t);
    }
    
    public String trace(MessageKey key) {
        return log(LogLevel.TRACE, key);
    }
 
    public String trace(MessageKey key, Throwable t) {
        return log(LogLevel.TRACE, key, t);
    }
 
    public String trace(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.TRACE, key, t, params);
    }
 
    public String trace(MessageKey key, Object... params) {
        return log(LogLevel.TRACE, key, params);
    }
    
    public String debug(MessageKey key) {
        return log(LogLevel.DEBUG, key);
    }
 
    public String debug(MessageKey key, Throwable t) {
        return log(LogLevel.DEBUG, key, t);
    }
 
    public String debug(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.DEBUG, key, t, params);
    }
 
    public String debug(MessageKey key, Object... params) {
        return log(LogLevel.DEBUG, key, params);
    }
 
    public String info(MessageKey key) {
        return log(LogLevel.INFO, key);
    }
 
    public String info(MessageKey key, Throwable t) {
        return log(LogLevel.INFO, key, t);
    }
 
    public String info(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.INFO, key, t, params);
    }
 
    public String info(MessageKey key, Object... params) {
        return log(LogLevel.INFO, key, params);
    }
    
    public String warn(MessageKey key) {
        return log(LogLevel.WARN, key);
    }
 
    public String warn(MessageKey key, Throwable t) {
        return log(LogLevel.WARN, key, t);
    }
 
    public String warn(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.WARN, key, t, params);
    }
 
    public String warn(MessageKey key, Object... params) {
        return log(LogLevel.WARN, key, params);
    }
    
    public String error(MessageKey key) {
        return log(LogLevel.ERROR, key);
    }
 
    public String error(MessageKey key, Throwable t) {
        return log(LogLevel.ERROR, key, t);
    }
 
    public String error(MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.ERROR, key, t, params);
    }
 
    public String error(MessageKey key, Object... params) {
        return log(LogLevel.ERROR, key, params);
    }
   
    public String log(String level, MessageKey key) {
        return log(LogLevel.parseLevel(level), key);
    }
    
    public String log(String level, MessageKey key, Throwable t) {
        return log(LogLevel.parseLevel(level), key, t);
    }
    
    public String log(String level, MessageKey key, Object... params) {
        return log(LogLevel.parseLevel(level), key, params);
    }
    
    public String log(String level, MessageKey key, Throwable t, Object... params) {
        return log(LogLevel.parseLevel(level), key, t, params);
    }
}
