/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/platform/linux/OperatingSystemImpl.java#1 $
 *
 * Copyright (c) 2006 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.platform.linux;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.netapp.common.MsgKey;
import com.netapp.common.execution.ShellCommand;
import com.netapp.common.execution.ShellRunnable;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;
import com.netapp.common.platform.DeviceAnalyzer;
import com.netapp.common.platform.DeviceType;
import com.netapp.common.platform.FileSystemMount;
import com.netapp.common.platform.PartitionTable;
import com.netapp.common.platform.PlatformException;
import com.netapp.common.platform.RawDeviceMapping;
import com.netapp.common.platform.linux.smbios.SmBiosParser;
import com.netapp.common.util.ConfigProperty;
import com.netapp.common.util.FileUtils;

/**
 * @author sirianni, sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class OperatingSystemImpl extends com.netapp.common.platform.unix.OperatingSystemImpl {
    protected static NALogger logger = NALogger.getLogger(OperatingSystemImpl.class);
    /**
     * Linux platforms that should use the O_DIRECT raw mechanism
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_DIRECTIO = ConfigProperty.stringProperty("versions.linux.directio");
    
    /**
     * Linux platforms that should use the RH4 MPIO handler
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_RH4MPIO = ConfigProperty.stringProperty("versions.linux.rh4mpio");
    
    /**
     * Linux platforms that should use the RH5 MPIO handler
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_RH5MPIO = ConfigProperty.stringProperty("versions.linux.rh5mpio");
    
    /**
     * Linux platforms that should use the SLESL 10 SP2 MPIO handler
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_SLES10MPIO  = ConfigProperty.stringProperty("versions.linux.sles10mpio");
    
    /**
     * Values of "os.arch" that should be considered x86
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_ARCH_X86 = ConfigProperty.stringProperty("versions.linux.arch.x86");
   
    /**
     * Values of "os.arch" that should be considered x64
     */
    protected static final ConfigProperty<String> VERSIONS_LINUX_ARCH_X64 = ConfigProperty.stringProperty("versions.linux.arch.x64");
    
    // Used internally; exposed for testing
    protected enum Distribution {
        UNKNOWN,
        REDHAT,
        ORACLE,
        SUSE
    }
    
    public static final String LINUX_PLATFORM_NAME = "linux";
   
    /**
     * /dev/sda1 /boot ext3 rw 0 0
     * jade-be.rtp.netapp.com:/vol/users/rross /u/rross nfs rw,bg,intr,hard,rsize=32768,wsize=32768,nfsvers=3,timeo=600,tcp,intr,addr=10.60.131.14 0 0
     */
    private static final Pattern MNTTAB_PATTERN = 
        Pattern.compile("(?:(\\S+):)?(\\S+) (\\S+) (\\S+) (\\S+) \\d+ \\d+.*");
    
    private static final int MNTTAP_SERVER_GROUP = 1;
    private static final int MNTTAB_DEVICE_GROUP = 2;
    private static final int MNTTAB_MOUNT_POINT_GROUP = 3;
    private static final int MNTTAB_TYPE_GROUP = 4;
    private static final int MNTTAB_OPTIONS_GROUP = 5;
    

    /** Linux Vendor */
    private Distribution distribution = Distribution.UNKNOWN;

    /** Major version */
    private int majorVersion;
    /** Minor version */
    private int minorVersion;
    
    /**
     *
     */
    public OperatingSystemImpl() {
        super();

        /* Identify which platform we're on. */
        File versionFile = null;

        try {
            // Check for OEL first; Otherwise it will detect as Red Hat
            if ((versionFile = new File("/etc/enterprise-release")).exists()) {
                BufferedReader in = new BufferedReader(new FileReader(versionFile));
                parseOracleLinuxVersion(in);
            }
            else if ((versionFile = new File("/etc/redhat-release")).exists()) {
                BufferedReader in = new BufferedReader(new FileReader(versionFile));
                parseRedhatVersion(in);
            }
            else if ((versionFile = new File("/etc/SuSE-release")).exists()) {
                BufferedReader in = new BufferedReader(new FileReader(versionFile));
                parseSuseVersion(in);
            }
        }
        catch (FileNotFoundException e) {
            // All the defaults are already set
        }

    }

    /**
     * Create an OperatingSystem impl for the supplied vendor and release file.
     * Used only for testing.
     * @param distribution
     * @param releaseString
     */
    OperatingSystemImpl(Distribution distribution, String releaseString) {
        super();
        BufferedReader in = new BufferedReader(new StringReader(releaseString));
        switch (distribution) {
            case REDHAT:
                parseRedhatVersion(in);
                break;
            case ORACLE:
                parseOracleLinuxVersion(in);
                break;
            case SUSE:
                parseSuseVersion(in);
                break;
            default:
        }
    }
    
    @Override
    public String getPlatformBrand() {
        MsgKey key;
        switch (this.distribution) {
            case REDHAT:
                key = MsgKey.PLATFORM_LINUX_REDHAT_BRAND;
                break;
            case ORACLE:
                key = MsgKey.PLATFORM_LINUX_ORACLE_BRAND;
                break;
            case SUSE:
                key = MsgKey.PLATFORM_LINUX_SUSE_BRAND;
                break;
            default:
                key = MsgKey.PLATFORM_LINUX_DEFAULT_BRAND;
        }
        return MessageBundle.localizeMessage(key);
    }
    
    @Override
    public String getArch() {
        String jvmArch = System.getProperty("os.arch");
        String x86 = VERSIONS_LINUX_ARCH_X86.lookup(this.getPlatformConfigBundle());
        String x64 = VERSIONS_LINUX_ARCH_X64.lookup(this.getPlatformConfigBundle());
        
        if (jvmArch.matches(x86)) {
            return "x86";
        }
        else if (jvmArch.matches(x64)) {
            return "x64";
        }
        return jvmArch;
    }
    
    /**
     * For historical reasons, this returns a combination of vendor and major distribution
     * version on Linux. This should not be changed without careful consideration. For example,
     * the strings "RHEL {X}" and "SEL {X}" are used throughout the diagnostics code to match
     * commands to run.
     * @see com.netapp.common.platform.OperatingSystem#getVersion()
     */
    @Override
    public String getVersion() {
        MsgKey key;
        switch (this.distribution) {
            case REDHAT:
                key = MsgKey.PLATFORM_LINUX_REDHAT_VERSION;
                break;
            case ORACLE:
                key = MsgKey.PLATFORM_LINUX_ORACLE_VERSION;
                break;
            case SUSE:
                key = MsgKey.PLATFORM_LINUX_SUSE_VERSION;
                break;
            default:
                key = MsgKey.PLATFORM_LINUX_DEFAULT_VERSION;
        }
        return MessageBundle.localizeMessage(key, this.majorVersion, this.minorVersion);
    }

    @Override
    public String getLibraryEnvironmentName() {
        return "LD_LIBRARY_PATH";
    }

    @Override
    public DeviceAnalyzer createDeviceAnalyzer() {
        LinuxDeviceAnalyzer analyzer = null;
        String currentVersion = getVersion();
        String currentFullVersion = currentVersion + "U" + this.minorVersion;
        logger.internal("Version: %s, FullVersion: $s", currentVersion, currentFullVersion);
        String directVersion = VERSIONS_LINUX_DIRECTIO.lookup(this.getPlatformConfigBundle());
        if (directVersion != null && currentVersion.matches(directVersion)) {
            analyzer = new DirectDeviceAnalyzer(this);
            logger.internal("Using O_DIRECT device analyzer");
        }
        else {
            analyzer = new RawDeviceAnalyzer(this);
            logger.internal("Using /dev/raw device analyzer");
        }
        
        analyzer.addBlockDeviceHandler(new ScsiDeviceHandler(this));
        String rh4mpio = VERSIONS_LINUX_RH4MPIO.lookup(this.getPlatformConfigBundle());
        String rh5mpio = VERSIONS_LINUX_RH5MPIO.lookup(this.getPlatformConfigBundle());
        
        String sles10mpio = VERSIONS_LINUX_SLES10MPIO.lookup(this.getPlatformConfigBundle());
        //First try matching Full version, if does not match then try using major version only
        if (rh4mpio != null && rh4mpio.indexOf(currentFullVersion + ",") >= 0) {
            analyzer.addBlockDeviceHandler(new RH4MultipathHandler(this));
            logger.internal("Using RHEL 4 MPIO handler for %s", currentFullVersion);
        }
        else if (rh5mpio != null && rh5mpio.indexOf(currentFullVersion + ",") >= 0) {
            analyzer.addBlockDeviceHandler(new RH5MultipathHandler(this));
            logger.internal("Using RHEL 5 MPIO handler for %s", currentFullVersion);
        }
        else if (rh4mpio != null && rh4mpio.indexOf(currentVersion + ",") >= 0) {
            analyzer.addBlockDeviceHandler(new RH4MultipathHandler(this));
            logger.internal("Using RHEL 4 MPIO handler");
        }
        else if (rh5mpio != null && rh5mpio.indexOf(currentVersion + ",") >= 0) {
            analyzer.addBlockDeviceHandler(new RH5MultipathHandler(this));
            logger.internal("Using RHEL 5 MPIO handler");
        }
        else if (sles10mpio != null && currentVersion.matches(sles10mpio)) {
            analyzer.addBlockDeviceHandler(new SLES10MultipathHandler(this));
            logger.internal("Using SLES 10 SP2 MPIO handler");
        }
        return analyzer;
    }
    
    /**
     * Parse the standard redhat-release file into a more standardized format
     * and initialize the vendor and version properties of this instance.
     * This usually looks like:
     * Red Hat Enterprise Linux AS release 3 (Taroon Update 4)
     *
     * @param f File to parse
     */
    protected void parseRedhatVersion(BufferedReader in) {
        parseRedhatDerivedVersion(Distribution.REDHAT, in);
    }
    
    protected void parseOracleLinuxVersion(BufferedReader in) {
        parseRedhatDerivedVersion(Distribution.ORACLE, in);
    }
    
    protected void parseRedhatDerivedVersion(Distribution distribution, BufferedReader in) {
        this.distribution = distribution;
        minorVersion = 0;
        majorVersion = 0;
        // This matches all of:
        // "Red Hat Enterprise Linux AS release 4 (Nahant Update 5)"
        // "Red Hat Enterprise Linux Server release 5 (Tikanga)"
        // "Red Hat Enterprise Linux Server release 5.2 (Tikanga)"
        // RHEL 3/4 will have the update version in group 3, RHEL 5 will use group 2
        Pattern p = Pattern.compile("elease (\\d+)(?:\\.(\\d+)|.*pdate (\\d+))?");
        try {
            String line = in.readLine();
            Matcher m = p.matcher(line);
            if (m.find()) {
                this.majorVersion = Integer.parseInt(m.group(1));
                if (m.group(2) != null) {
                    this.minorVersion = Integer.parseInt(m.group(2));
                }
                else if (m.group(3) != null) {
                    this.minorVersion = Integer.parseInt(m.group(3));
                }
            }
            in.close();
        }
        catch (IOException e) {

        }
    }

    /**
     * Parse the standard SuSE-release file into our standard format.
     */
    protected void parseSuseVersion(BufferedReader in) {
        distribution = Distribution.SUSE;
        minorVersion = 0;
        majorVersion = 0;
        try {
            // Ignore the first line
            in.readLine();

            // Second line has the version number on it
            String majorLine = in.readLine();
            String minorLine = in.readLine();
            
            Pattern majorPattern = Pattern.compile("VERSION = (\\d+)");
            Pattern minorPattern = Pattern.compile("PATCHLEVEL = (\\d+)");
            
            Matcher m = majorPattern.matcher(majorLine);
            if (m.matches()) {
                majorVersion = Integer.parseInt(m.group(1));
                
                if (minorLine != null) {
                    m = minorPattern.matcher(minorLine);
                    if (m.matches()) { 
                        minorVersion = Integer.parseInt(m.group(1));
                    }
                }
            }
        }
        catch (IOException e) {
        }
    }

    public RawDevicesFile getRawDevicesFile() {
        switch (this.distribution) {
            case REDHAT:
            case ORACLE:
                return new RedhatRawDevicesFile();
            case SUSE:
                return new SuseRawDevicesFile();
            default:
                assert false : "Unknown vendor " + this.distribution;
                return null;
        }
    }

    @Override
    protected ShellRunnable createDFCommand(List<File> fileList) {
        ShellCommand shellCommand = new ShellCommand();
        shellCommand.append("df", "-PT");
        for (File file : fileList) {
            shellCommand.append(file.getAbsolutePath());
        }
        return shellCommand;
    }

    @Override
    protected DFResultConstructor createDFResultConstructor() {
        return new com.netapp.common.platform.linux.DFResultConstructor();
    }
    
    @Override
    public Collection<FileSystemMount> getMountedFilesystems() {
        List<Matcher> mnttab = null;
        try {
            mnttab = FileUtils.matchLines(MNTTAB_PATTERN, getMountTableFile());
        }
        catch (IOException e) {
            throw new PlatformException(
                logger.debug(MsgKey.PLATFORM_MOUNT_TABLE_READ_FAILED, e.getLocalizedMessage()),
                e);
        }
        
        List<FileSystemMount> results = new ArrayList<FileSystemMount>();
        for (Matcher m : mnttab) { 
            String type = m.group(MNTTAB_TYPE_GROUP);
            if ("autofs".equals(type)) {
                continue;
            }
            String server = m.group(MNTTAP_SERVER_GROUP);
            String device = m.group(MNTTAB_DEVICE_GROUP);
            String options = m.group(MNTTAB_OPTIONS_GROUP);
            String mountPoint = m.group(MNTTAB_MOUNT_POINT_GROUP);

            results.add(new FileSystemMount(server, device, mountPoint, type, options));
        }
        return results;
    }
    
    /**
     * Unit test hook
     */
    protected File getMountTableFile() {
        return new File("/etc/mtab");
    }
    
    /**
     * Overridden to support copying raw devices on Linux.
     * 
     * Copying a raw device (eg. /dev/raw/raw1) using streams
     * results in an IOException with the message "Invalid Argument".
     * (Note the same message results when using the 'cp' os command).
     * 
     * Copying a raw device using a {@link java.io.RandomAccessFile#getChannel()} as input
     * and a {@link java.nio.channels.FileChannel} as output, results in all but the last
     * block of the device copied, and then an IOException with the message
     * "No such device or address".  
     * (Note the same message results when using the 'dd' os command).
     * 
     * Also, the raw device file size is reported as 0, so the
     * {@link #copyFileUsingChannels(File, File)} method cannot be used.
     * 
     * Therefore, to support copying raw devices on linux, instead of
     * copying the raw device (eg. /dev/raw/raw2) directly, the actual
     * disk partition (eg. /dev/sdf1) is copied.
     * 
     */
    @Override
    protected void copyFileUsingStreams(File src, File dest) throws IOException {
        
        DeviceAnalyzer deviceAnalyzer = createDeviceAnalyzer();
        if (deviceAnalyzer.determineDeviceType(src.getCanonicalFile()) == DeviceType.RAW_DEVICE) {
            RawDeviceMapping srcRawDeviceMapping = deviceAnalyzer.lookupMappingByRawDevice(src);
            src = srcRawDeviceMapping.getDisk();
            
            if (deviceAnalyzer.determineDeviceType(dest.getCanonicalFile()) == DeviceType.RAW_DEVICE) {
                RawDeviceMapping destRawDeviceMapping = deviceAnalyzer.lookupMappingByRawDevice(dest.getCanonicalFile());
                dest = destRawDeviceMapping.getDisk();
            }
        }
        super.copyFileUsingStreams(src, dest);
    }

    @Override
    public PartitionTable readPartitionTable(File device) {
        try {
            return new DOSPartitionTable(device, this);
        }
        catch (IOException e) {
            throw new PlatformException(e);
        }
    }
    
    /**
     * Throws a PlatformException if an error occurs while reading the UUID.
     */
    @Override
    public String getBiosUuid() {
        try {
            return SmBiosParser.readUuid();
        }
        catch (SecurityException e) {
            throw new PlatformException(logger.error(MsgKey.PLATFORM_OS_UUID_NOT_FOUND, e));
        }
        catch (IOException e) {
            throw new PlatformException(logger.error(MsgKey.PLATFORM_OS_UUID_NOT_FOUND, e));
        }
    }

    @Override
    public String toString() {
        MsgKey key;
        switch (this.distribution) {
            case REDHAT:
            case ORACLE:
            case SUSE:
                key = MsgKey.PLATFORM_LINUX_BRANDED_DESCRIPTION;
                break;
            case UNKNOWN:
            default:
                key = MsgKey.PLATFORM_LINUX_UNKNOWN_DESCRIPTION;
        }
        return MessageBundle.localizeMessage(key,
            this.getPlatformBrand(),
            this.getArch(),
            this.majorVersion,
            this.minorVersion);
    }
}
