/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/server/ServerFacade.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.server;

import com.netapp.common.MsgKey;
import com.netapp.common.cli.CliCommand;
import com.netapp.common.cli.CliParameter;
import com.netapp.common.cli.CliParser;
import com.netapp.common.cli.Command;
import com.netapp.common.logging.MessageBundle;
import com.netapp.common.logging.NALogger;

public class ServerFacade {
    private static final NALogger LOGGER = NALogger.getLogger(ServerFacade.class);
    
    public static final String COMMAND_NAME_START = "start";
    public static final String COMMAND_NAME_STOP = "stop";
    public static final String COMMAND_NAME_STATUS = "status";
    
    public static final String DEFAULT_CONFIGURATION_FILE = "classpath:conf/jetty.xml";
    public static final String DEFAULT_COMMAND_NAME = "server";
    
    
    private String configurationFile;
    private String beanName;
    private String serverName;
    private String commandName;
    private int port;
    private AbstractJettyServer jettyServer;
    
    public ServerFacade() {
        this(DEFAULT_CONFIGURATION_FILE);
    }
    
    public ServerFacade(String configurationFile) {
        this(new JettyServer(configurationFile));
        setConfigurationFile(configurationFile);
    }
    
    public ServerFacade(AbstractJettyServer jettyServer) {
        this.jettyServer = jettyServer;
        
        setBeanName(JettyServer.DEFAULT_BEAN_NAME);
        setServerName(JettyServer.DEFAULT_SERVER_NAME);
        setCommandName(DEFAULT_COMMAND_NAME);
        
        this.jettyServer.setBeanName(getBeanName());
        this.jettyServer.setServerName(getServerName());
    }

    @CliCommand(command = COMMAND_NAME_START)
    public void start() {        
        if (-1 != getPort()) { // allow port override when starting
            this.jettyServer.setPort(getPort());
        }
        
        this.jettyServer.start();
        
        String commonDir = System.getProperty("common.dir");
        
        if (null != commonDir) {
            
            this.jettyServer.getDescriptor().setPidDirectory(commonDir + ServerDescriptor.DEFAULT_PID_DIR.getPath());
            LOGGER.internal("Using directory '%s' for storing pid files.", this.jettyServer.getDescriptor().getPidDirectory());
        }
        
        this.jettyServer.getDescriptor().toFile();
        System.out.println(MessageBundle.localizeMessage(MsgKey.SERVER_JETTY_SERVER_STARTED, this.jettyServer.getPort()));
    }
    
    @CliCommand(command = COMMAND_NAME_STOP)
    public void stop() {
        if (-1 != getPort()) { // allow port override when stopping
            this.jettyServer.setPort(getPort());
        }
        
        String commonDir = System.getProperty("common.dir");
        
        if (null != commonDir) {
            this.jettyServer.setDescriptor(new ServerDescriptor(getServerName(), 0, 0)); // need an empty one
            this.jettyServer.getDescriptor().setPidDirectory(commonDir + ServerDescriptor.DEFAULT_PID_DIR.getPath());
            LOGGER.internal("Using directory '%s' for retrieving pid files.", this.jettyServer.getDescriptor().getPidDirectory());
        }
        
        this.jettyServer.stop();
        this.jettyServer.getDescriptor().deleteFile(); // clean up after ourself
        System.out.println(MessageBundle.localizeMessage(MsgKey.SERVER_JETTY_SERVER_STOPPED, this.jettyServer.getPort()));
    }
    
    @CliCommand(command = COMMAND_NAME_STATUS)
    public void status() {
        if (-1 != getPort()) { // allow port override when stopping
            this.jettyServer.setPort(getPort());
        }

        String commonDir = System.getProperty("common.dir");
        
        if (null != commonDir) {
            ServerDescriptor dummyDescriptor = new ServerDescriptor(getServerName(), 0, 0);
            dummyDescriptor.setPidDirectory(commonDir + ServerDescriptor.DEFAULT_PID_DIR.getPath());
            LOGGER.internal("Using directory '%s' for retrieving pid files.", dummyDescriptor.getPidDirectory());

            this.jettyServer.setDescriptor(getActiveServerDescriptor(dummyDescriptor, this.jettyServer));
            this.jettyServer.getDescriptor().setPidDirectory(commonDir + ServerDescriptor.DEFAULT_PID_DIR.getPath());

        }
        
        System.out.println(MessageBundle.localizeMessage(MsgKey.SERVER_JETTY_SERVER_STATUS, this.jettyServer.getDescriptor().getPid()));
    }
    
    @CliCommand(command = Command.COMMAND_NAME_HELP)
    public void help() {
        CliParser cliParser = new CliParser(this.getClass());
        String usage = cliParser.getUsage(getCommandName());
        System.out.println(usage);
    }

    public String getConfigurationFile() {
        return configurationFile;
    }

    public void setConfigurationFile(String webAppLocation) {
        this.configurationFile = webAppLocation;
    }

    public String getBeanName() {
        return beanName;
    }

    public void setBeanName(String beanName) {
        this.beanName = beanName;
    }

    public String getServerName() {
        return serverName;
    }

    public void setServerName(String serverName) {
        this.serverName = serverName;
    }

    public String getCommandName() {
        return commandName;
    }

    public void setCommandName(String commandName) {
        this.commandName = commandName;
    }

    public int getPort() {
        return this.port;
    }

    @CliParameter(option = "port")
    public void setPort(Integer port) {
        if (null == port) {
            this.port = -1;
        }
        else {
            this.port = port;
        }
    }
    
    /**
     * If the server is running, then this method returns an instance of ServerDescriptor.
     * If the SMO server is NOT running, then this method
     *    a) deletes the descriptor file (if it exists), and
     *    b) returns null.
     */
    private ServerDescriptor getActiveServerDescriptor(ServerDescriptor dummyDescriptor, AbstractJettyServer server) {
        server.loadConfiguration();
        for (ServerDescriptor ssd : dummyDescriptor.getAllDescriptors()) {
            LOGGER.internal("Found PID file " + ssd.getFileName());
            LOGGER.internal("PID file content" + ssd.toString());
            LOGGER.internal("Requested Port " + server.getPort());

            if ((ssd.getPort() == server.getPort())) {
//                LOGGER.internal("Checking if Server is running on descriptor file port " + ssd.getPort());
//                ServerStatusResult statusResult = this.getServerStatus(ssd.getPort());
//                if (statusResult != null) {
                LOGGER.internal("Server is running on port " + ssd.getPort() + " with pid " + ssd.getPid());
                return ssd;
//                }
//                else {
//                    LOGGER.internal("Deleting stale descriptor file " + ssd.getFileName());
//                    ssd.deleteFile();
//                }
            }
        }
        return null;
    }
    
}
