/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/ChangeTracker.java#1 $
 * 
 * Copyright (c) 2008 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.util.LinkedList;
import java.util.Timer;
import java.util.TimerTask;

import com.netapp.common.logging.NALogger;

/**
 * Maintain a time-limited list of change notifications and allow clients
 * to retrieve all updates since a previous change set.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ChangeTracker<T> {
    private static final NALogger logger = NALogger.getLogger(ChangeTracker.class);
    private static final Timer SCAVENGER_TIMER = new Timer("ChangeTrackerScavenger", true);
    
    private long nextChangeNumber = 1;
    private Generation head;
    private TimerTask scavengerTask;

    /**
     * Create a ChangeTracker with the default object lifetime
     */
    public ChangeTracker() {
        this(1000 * 60 * 10); // 10m
    }

    /**
     * Create a new ChangeTracker with the specified object lifetime
     * @param maxLifetime The lifetime of objects added to this tracker, in ms.
     */
    public ChangeTracker(long maxLifetime) {
        this.scavengerTask = new ScavengerTask();
        SCAVENGER_TIMER.schedule(this.scavengerTask, maxLifetime, maxLifetime);
    }

    /**
     * Post a new notification to this tracker
     * @param change
     */
    public synchronized void addChange(T change) {
        // TODO: Use the cmpx support in java.util.concurrent to do this lockless
        Generation g = new Generation(change, this.head);
        this.head = g;
        logger.internal("Created change %s", g.getId());
    }

    /**
     * Retrieve all changes since the given tag.
     * @param since The tag from a previous call to this method, or 0 if this is the first invocation.
     * @return A change set with the given changes and the next tag
     */
    public ChangeSet<T> getChanges(long since) {
        return getChanges(since, null);
    }

    /**
     * Retrieve all changes since the given tag, filtered by the given predicate
     * @param since The tag from the previous call to this method, or 0 if this is the first invocation.
     * @param predicate A predicate to filter results with. Only changes which this predicate evaluates to true will be included.
     * @return A change set with the given changes and the next tag
     */
    public ChangeSet<T> getChanges(long since, Predicate<T> predicate) {
        LinkedList<T> results = new LinkedList<T>();

        Generation c = this.head;
        long tag = (c != null) ? c.getId() : 0;
        while (c != null && c.getId() > since) {
            if (predicate == null || predicate.evaluate(c.getValue())) {
                // We want the oldest results up front.
                results.addFirst(c.getValue());
            }
            c = c.getPrev();
        }
        
        //logger.internal("Returning %s changes since tag %s", results.size(), since);
        return new ChangeSet<T>(tag, results);
    }

    @Override
    public void finalize() throws Throwable {
        if (this.scavengerTask != null) {
            this.scavengerTask.cancel();
        }
        super.finalize();
    }
    
    /**
     * This holds both our generational information and also
     * serves as a linked list. The util lists don't give us
     * a quick Truncate option.
     */
    private class Generation {
        private final long id; 
        private final long time;
        private final T value;

        private Generation prev;

        Generation(T value, Generation prev) {
            // This is always called from within the synchronized block of
            // addChange. Otherwise we need to make this an AtomicLong.
            this.id = ChangeTracker.this.nextChangeNumber++;
            this.time = System.currentTimeMillis();
            this.value = value;
            this.prev = prev;
        }

        public long getId() {
            return this.id;
        }

        public long getTime() {
            return this.time;
        }

        public T getValue() {
            return this.value;
        }

        public Generation getPrev() {
            return this.prev;
        }

        public long truncate() {
            this.prev = null;
            return this.time;
        }
    }
    
    private class ScavengerTask extends TimerTask {
        private Generation last;
        @Override
        public void run() {
            if (last != null) {
                last.truncate();
            }
            last = ChangeTracker.this.head;
        }
        
    }
}
