/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/ConfigBundle.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Collection;
import java.util.Properties;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * Maintain a list of application configuration values.
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class ConfigBundle {
    private static final NALogger logger = NALogger.getLogger(ConfigBundle.class);
    private Properties properties;
    private ConfigProperty<File> installDirProperty = new ConfigProperty<File>(File.class, "install.dir");
    private File installDir;
    
    /**
     * Construct an empty ConfigBundle.
     */
    public ConfigBundle() {
        this.properties = new Properties();
    }
    
    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supercede definitions from earlier files.
     * @param configFiles
     */
    public ConfigBundle(String... configFiles) {
        this(Thread.currentThread().getContextClassLoader(), configFiles);
    }
    
    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     * @param classLoader
     */
    public ConfigBundle(ClassLoader classLoader, String... configFiles) {
        this(load(classLoader, configFiles));
    }
    
    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     */
    public ConfigBundle(Collection<String> configFiles) {
        this(Thread.currentThread().getContextClassLoader(), configFiles);
    }
    
    /**
     * Construct a new Configuration Bundle from the supplied file names. All files are relative to the executing class path.
     * Later files will supersede definitions from earlier files.
     * @param configFiles
     * @param classLoader
     */
    public ConfigBundle(ClassLoader classLoader, Collection<String> configFiles) {
        this(load(classLoader, configFiles.toArray(new String[]{})));
    }

    public ConfigBundle(Properties properties) {
        this.properties = properties;
    }
    
    /**
     * Construct a new Configuration Bundle by combining bundles.
     * Later bundles supersede definitions from earlier ones.
     * @param bundles
     */
    public ConfigBundle(ConfigBundle... bundles) {
        this.properties = new Properties();
        for (ConfigBundle bundle : bundles) {
            if (bundle != null) {
                this.properties.putAll(bundle.properties);
            }
        }
    }
    
    public File getInstallDir() {
        if (this.installDir == null) {
            this.installDir = installDirProperty.lookup(this);
        }
        
        if (this.installDir == null) {
            String env = System.getenv("INSTALL_DIR");
            if (env != null) {
                this.installDir = new File(env);
            }
        }
        
        if (this.installDir == null) {
            this.installDir = new File(".");
        }
        return installDir;
    }
    
    public String getProperty(String key) {
        return this.properties.getProperty(key);
    }
    
    public void setProperty(String key, String value) {
        this.properties.setProperty(key, value);
    }
    
    public boolean containsKey(String key) {
        return this.properties.containsKey(key);
    }
    
    protected Properties getProperties() {
        return this.properties;
    }
    
    public static Properties load(ClassLoader loader, String[] configFiles) {
        String file = null;
        Properties returnProps = new Properties();
        try {    
            for (String ifile : configFiles) {
                file = ifile;
                InputStream stream = loader.getResourceAsStream(file);
                
                if (stream != null) {
                    Properties props = new Properties(returnProps);
                    props.load(stream);
                    stream.close();
                    returnProps = props;
                }
            }
        }
        catch (IOException e) {
            throw new RuntimeException(logger.error(MsgKey.UTIL_CONFIG_EXCEPTION_LOADING_FILE, e, file, e.getLocalizedMessage()));
        }
        return returnProps;
    }
}
