/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/util/Utils.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.util;

import java.io.File;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.UnknownHostException;
import java.text.Format;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.Enumeration;
import java.util.GregorianCalendar;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.apache.commons.lang.builder.ReflectionToStringBuilder;
import org.apache.commons.lang.builder.ToStringStyle;

import com.netapp.common.MsgKey;
import com.netapp.common.logging.NALogger;

/**
 * General utility routines.
 * Functions dealing primarily with manipulating Classes, Strings or Java Collections should be
 * placed in ClassUtils, StringUtils or CollectionUtils respectively.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class Utils {
    
    private static final NALogger logger = NALogger.getLogger(Utils.class);
    
    private static final NumberFormat NUMBER_FORMAT = NumberFormat.getInstance();

    static {
        NUMBER_FORMAT.setGroupingUsed(false);
    }

    protected Utils() {
    }

    
    public static String nullSafeString(Object o) {
        return o == null ? "" : o.toString();
    }
    
    public static String prettyString(Throwable t) {
        return t.getLocalizedMessage() != null ? t.getLocalizedMessage() : t.toString();
    }

    public static String reflectiveToString(Object o) {
        if (o == null) {
            return "null";
        }
        
        if (o instanceof String) {
            return "String[value=" + (String) o + "]";
        }
        
        ToStringStyle toStringStyle = new ToStringStyle() {
            {
                this.setUseShortClassName(true);
                this.setUseIdentityHashCode(false);
            }
            @Override
            public void append(StringBuffer buffer, String fieldName, Object fieldValue, Boolean fullDetail) {
                if (fieldValue instanceof String || fieldValue instanceof Integer 
                        || fieldValue instanceof Boolean || fieldValue instanceof Long) {
                    super.append(buffer, fieldName, fieldValue, fullDetail);
                }
                else {
                    super.append(buffer, fieldName, 
                        Utils.reflectiveToString(fieldValue, this), fullDetail);
                }
            }
        };
        
        return reflectiveToString(o, toStringStyle);
    }

    private static String reflectiveToString(Object object, ToStringStyle toStringStyle) {
        ReflectionToStringBuilder reflectionToStringBuilder = new ReflectionToStringBuilder(object,
            toStringStyle,
            null,         // buffer
            Object.class, // up to class
            false,        // transients?
            false);       // statics?

        reflectionToStringBuilder.setExcludeFieldNames(new String[] {"password"});
        
        return reflectionToStringBuilder.toString();
    }

    /**
     * Sleep for the specified number of milliseconds without throwing an 
     * InterruptedException.
     * @param milliseconds The number of milliseconds to sleep for
     */
    public static void sleep(long milliseconds) {
        try {
            Thread.sleep(milliseconds);
        }
        catch (InterruptedException e) {
            // do nothing
        }
    }

    /**
     * Read a 32 bit unsigned value in little endian format
     */
    public static long unsigned32LE(byte[] b, int offset) {
        return ((0xFF & b[offset + 0]) << 0)
             | ((0xFF & b[offset + 1]) << 8)
             | ((0xFF & b[offset + 2]) << 16)
             | ((0xFF & b[offset + 3]) << 24);
    }

    /**
     * Read a 16 bit unsigned value in little endian format
     */
    public static int unsigned16LE(byte[] b, int offset) {
        return ((0xFF & b[offset + 0]) << 0)
             | ((0xFF & b[offset + 1]) << 8);
    }
    
    /**
     * Read a 32 bit unsigned value in big endian format
     */
    public static long unsigned32BE(byte[] b, int offset) {
        return ((0xFF & b[offset + 0]) << 24)
             | ((0xFF & b[offset + 1]) << 16)
             | ((0xFF & b[offset + 2]) << 8)
             | ((0xFF & b[offset + 3]) << 0);
    }
    
    /**
     * Read a 16 bit unsigned value in big endian format
     */
    public static int unsigned16BE(byte[] b, int offset) {
        return ((0xFF & b[offset + 0]) << 8)
             | ((0xFF & b[offset + 1]) << 0);
    }
    
    public static short unsigned8(byte b) {
        return (short) (0xFF & b);
    }

    public static String byteArrayToHexString(byte[] byteArray, int offset, int length) {
        return byteArrayToHexString(byteArray, 0, offset, length);
    }
    
    public static String byteArrayToHexString(byte[] byteArray) {
        return byteArrayToHexString(byteArray, 0, 0, byteArray.length);
    }
    
    public static String byteArrayToHexString(byte[] byteArray, int grouping) {
        return byteArrayToHexString(byteArray, grouping, 0, byteArray.length);
    }
    
    /**
     * @param grouping If greater than 0, will be used as the number of bytes per space-delimited
     * character group in the hex string 
     */
    public static String byteArrayToHexString(byte[] byteArray, int grouping, int offset, int length) {
        StringBuffer hexString = new StringBuffer(byteArray.length);
        for (int i = offset; i < offset + length; i++) {
            if (grouping > 0 && i != 0 && i % grouping == 0) {
                hexString.append(" ");
            }
        
            String hex = Integer.toHexString(0x0100 + (byteArray[i] & 0x00FF)).substring(1);
            hexString.append((hex.length() < 2 ? "0" : "") + hex);
        }
        return hexString.toString();
    }

    public static String byteToHexString(byte b) {
        String hex = Integer.toHexString(0x0100 + (b & 0x00FF)).substring(1);
        return (hex.length() < 2 ? "0" : "") + hex;
    }
    
    /**
     * Convert a number to a string (without commas)
     * @param number A number (e.g 1234)
     * @return A string representing that number without commas (e.g. 1234 not 1,234)
     */
    public static String numberToStringNoCommas(Object number) {
        return NUMBER_FORMAT.format(number);
    }
    
    /**
     * Get the name for this host
     * @return The host name for this machine, or "localhost" if that cannot be determined.
     */
    public static String getLocalHostName() {
        try {
            return InetAddress.getLocalHost().getHostName();
        }
        catch (UnknownHostException e) {
            return "localhost";
        }   
    }

    /**
     * Get the fully qualified host name for this host.
     * @return The fully qualified host name for this machine, or "localhost" if that cannot be determined.
     */
    public static String getCanonicalLocalHostName() {
        try {
            return InetAddress.getLocalHost().getCanonicalHostName();
        }
        catch (UnknownHostException e) {
            return "localhost";
        }   
    }
    
    public static List<InetAddress> getLocalHostAddresses() throws SocketException {
        List<InetAddress> list = new ArrayList<InetAddress>();
 
        Enumeration<NetworkInterface> interfaces = NetworkInterface.getNetworkInterfaces();
        if (interfaces != null) {
            while (interfaces.hasMoreElements()) {
                NetworkInterface networkInterface = (NetworkInterface) interfaces.nextElement();
                Enumeration<InetAddress> addresses = networkInterface.getInetAddresses();
                if (addresses != null) {
                    while (addresses.hasMoreElements()) {
                        InetAddress address = (InetAddress) addresses.nextElement();
                        list.add(address);
                    }
                }
            }
        }
 
        return list;
    }
    
    /**
     * @param host The hostname to check
     * @return true if the hostname maps to the localhost
     */
    public static boolean isLocalhost(String host) {
        
        try {
            for (InetAddress localHostAddress : getLocalHostAddresses()) {
                if (hostEquals(localHostAddress, host)) {
                    return true;
                }
            }
            return false;
        }
        catch (SocketException e) {
            logger.warn(MsgKey.UTIL_LOCAL_HOST_ADDRESSES, e);
            return false;
        }
    }
    
    /**
     * @param addr1 An internet address object
     * @param host2 A short-form hostname, fully-qualified hostname, or IP address.
     * @return true if the address and host string map to the same physical machine  
     */
    public static boolean hostEquals(InetAddress addr1, String host2) {
        try {
            InetAddress addr2 = InetAddress.getByName(host2);
            boolean areEquals = addr1.equals(addr2);
            logger.internalTrace("Utils.hostEquals(" + addr1 + ", " + host2 + " ["
                    + addr2.getHostAddress() + "]) = " + Boolean.toString(areEquals));
            return areEquals;
        }
        catch (UnknownHostException e) {
            logger.warn(MsgKey.UTIL_UNKNOWN_HOST, e.getLocalizedMessage());
            return false;
        }
    }
    
    /**
     * @param host1 A short-form hostname, fully-qualified hostname, or IP address.
     * @param host2 A short-form hostname, fully-qualified hostname, or IP address.
     * @return true if the two host strings map to the same physical machine      
     */
    public static boolean hostEquals(String host1, String host2) {
        if (host1.equals(host2)) {
            return true;
        }
        else {
            try {
                InetAddress addr1 = InetAddress.getByName(host1);
                return hostEquals(addr1, host2);
            }
            catch (UnknownHostException e) {
                logger.warn(MsgKey.UTIL_UNKNOWN_HOST, e.getLocalizedMessage());
                return false;
            }
        }
    }

    /**
     * Return true if s is null or completely whitespace.
     * @deprecated Use StringUtils.isBlank() instead.
     */
    @Deprecated
    public static boolean isEmpty(String s) {
        return StringUtils.isBlank(s);
    }
    
    /**
     * @see CollectionUtils#isEmpty(Collection)
     * @param c The Collection instance to check for emptiness
     * @return Returns <code>true</code> if null or has a size of 0;  else, <code>false</code>
     * @deprecated
     */
    public static boolean isEmpty(Collection<?> c) {
        return CollectionUtils.isEmpty(c);
    }

    /**
     * @see CollectionUtils#isEmpty(Map)
     * @param m The Map instance to check for emptiness
     * @return Returns <code>true</code> if null else <code>m.isEmpty()</code>
     * @deprecated
     */
    public static boolean isEmpty(Map<?, ?> m) {
        return CollectionUtils.isEmpty(m);
    }

    /**
     * @see CollectionUtils#isEmpty(Object[])
     * @deprecated
     */
    public static boolean isEmpty(Object[] a) {
        return CollectionUtils.isEmpty(a);
    }
    
    /**
     * @see StringUtils#findRegExMatches(String, CharSequence)
     * @deprecated
     */
    @Deprecated
    public static String[] findRegExMatches(String patternStr, CharSequence input) {
        return StringUtils.findRegExMatches(patternStr, input);
    }
    
    /**
     * @see StringUtils#containsAny(String, String...)
     * @deprecated
     */
    @Deprecated
    public static boolean containsAny(String s, String... substrings) {
        return StringUtils.containsAny(s, substrings);
    }
    
    /**
     * Returns String representation of the current date using the format specified.
     */
    public static String getCurrentDateUsingFormat(String format) {
        Format formatter = new SimpleDateFormat(format);
        return formatter.format(new Date());
    }
    
    /**
     * Gets the file name from the given file path. It first tries the system
     * based delimiter. If that fails, it then tries to use a unix based 
     * delimiter. If that fails, it tries to use a windows based delimiter. 
     * If THAT fails, it returns the file path.
     * 
     * @param filePath the given file path
     * @return the file name
     */
    public static String getFileName(String filePath) {

        if (filePath == null) {
            throw new IllegalArgumentException("File path cannot be null.");
        }

        String retVal = null;

        /*
         * first try system defined delimiter
         */
        retVal = getEndSubstring(filePath, System.getProperty("file.separator"));

        if (retVal != null) {
            return retVal;
        }

        /*
         * next try unix based delimiter
         */
        retVal = getEndSubstring(filePath, "/");

        if (retVal != null) {
            return retVal;
        }

        /*
         * finally try windows based delimiter
         */
        retVal = getEndSubstring(filePath, "\\");

        if (retVal != null) {
            return retVal;
        }

        return filePath;
    }
    
    /**
     * Returns the substring of str1 after str2. If str2 is not found,
     * null is returned. 
     *
     * @param str1 the string to look in
     * @param str2 the string to look for
     * @return the substring of str1 after str2, or null if str2 is not found
     */
    private static String getEndSubstring(String str1, String str2) {

        int ndx = str1.lastIndexOf(str2);

        if (ndx >= 0) {
            return str1.substring(ndx + 1);
        }

        return null;
    }
    
    /**
     * returns true if and only if lowval <= i <= highval
     */
    public static boolean integerInRange(int i, int lowval, int highval) {
        return ((i >= lowval) && (i <= highval));
    }
    
    /**
     * Creates a Calendar object using the Date specified.
     */
    public static Calendar calendarFromDate(Date d) {
        if (d == null) {
            return null;
        }
        
        Calendar calendar = new GregorianCalendar();
        calendar.setTime(d);
        return calendar;
    }
    
    
    /**
     * Returns a boolean to indicate if a given number is between a low value
     * and a high value, inclusive.
     * @param value The input number (as a String).  (If the String cannot be parsed
     * to an integer, then the method returns false.)
     * @param lowval The low value.
     * @param highval The high value.
     * @return A boolean.
     * @throws NumberFormatException if value is not numeric.
     */    
    public static boolean integerInRange(String value, int lowval, int highval) {
        if (StringUtils.isEmpty(value)) {
            return false;
        }
        try {
            int i = Integer.parseInt(value);
            return Utils.integerInRange(i, lowval, highval);
        }
        catch (NumberFormatException nfe) {
            if (Utils.isNumeric(value)) {
                // The value is not a 32-bit integer, but it is a number. 
                // Clearly the value is not in the range of two integers.
                return false;
            }
            else {
                throw nfe;
            }
        }
    }
    
    
    /**
     * Check if the supplied file exists and is a directory.
     * 
     * @param file
     * @return true if file is a directory, false if not
     */
    public static boolean isDirectory(File file) {
        boolean isDirectory = false;
        
        if (file.exists() && file.isDirectory()) {
            isDirectory = true;
        }
        
        return isDirectory;
    }
    
    
    /**
     * Returns true if one and only one of the boolean parameters is true.
     * 
     * @return true if one and only one of the boolean parameters is true.
     */
    public static boolean isExactlyOneTrue(boolean... b) {
        boolean found = false;
        for (boolean testBoolean : b) {
            if (testBoolean) {
                if (found) {
                    return false;
                }
                found = true;
            }
        }
        return found;
    }
    
    
    /**
     * Returns true only if all of the following conditions are met
     * - arg1 is not null
     * - arg2 is not null
     * - arg1 and arg2 are equal according to the "equals" method
     * @param arg1
     * @param arg2
     * @return
     */
    public static boolean equalAndRealValues(Object arg1, Object arg2) {
        if ((arg1 != null) && (arg2 != null)) {
            return arg1.equals(arg2);
        }
        else {
            return false;
        }
    }
    
    private static final Pattern NUMBER_PATTERN = Pattern.compile("[+-]?[0-9]+");

    /**
     * Given a String, thie method returns a boolean to indicate if the String
     * is a valid positive or negative integer.  Values "55", "+55", "-55", and
     * "812345678912345" (which cannot be parsed by java.lang.Integer) all 
     * return true.  Values "55.0", "5A", and "" all return false.
     * @param value - String containing a value.
     * @return boolean
     */
    public static boolean isNumeric(String value) {
        Matcher myMatcher = NUMBER_PATTERN.matcher(value);
        return myMatcher.matches();
    }
    
    /**
     * Returns a localized message if it exists. If it does not exist then t.toString is returned.
     */
    public static String niceThrowableMessage(Throwable t) {
        return t.getLocalizedMessage() != null ? t.getLocalizedMessage() : t.toString();
    }
    
    /**
     * Returns a localized message if it exists along with the stacktrace.
     * If the localized message does not exist then t.toString is returned (with the stacktrace).
     */
    public static String niceThrowableMessageWithStacktrace(Throwable t) {
        String msg = Utils.niceThrowableMessage(t);
        StringWriter out = new StringWriter();
        t.printStackTrace(new PrintWriter(out));
        String newline = System.getProperty("line.separator");
        return msg + newline + out;
    }
}
