/*
 * $Id: //depot/prod/newport_news/Radamant.1/src/java/com/netapp/common/zapi/generator/autozapi/AutoZapiGenerator.java#1 $
 * 
 * Copyright (c) 2007 Network Appliance, Inc.
 * All rights reserved.
 */
package com.netapp.common.zapi.generator.autozapi;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import com.netapp.common.Version;
import com.netapp.common.generator.NetappCodeFormatter;
import com.netapp.common.logging.NALogger;
import com.netapp.common.util.ClassUtils;
import com.netapp.common.zapi.generator.GeneratorUtils;
import com.netapp.common.zapi.parser.Namespace;
import com.netapp.common.zapi.parser.ZapidocParser;
import java.util.Properties;

/**
 * Parse a supplied zapidoc file into a set of annotated Api and Typedef classes
 * for use with {@link com.netapp.common.zapi.ZAPIRunner}.
 * 
 * @author sunday
 * @version $Revision: #1 $ $Date: 2010/08/20 $
 */
public class AutoZapiGenerator {
    static final NALogger logger = NALogger.getLogger(AutoZapiGenerator.class);
    
    public static final NetappCodeFormatter FORMATTER;
    static {
        FORMATTER = new NetappCodeFormatter();
        FORMATTER.addHeaderLine("Generated by zapi2java " + Version.VERSION + " (" + Version.CHANGEID + ")");
    }
    
    private final String packageName;
    
    private final ZapidocParser parser;

    private AutoZapiGeneratorFactory generatorFactory;
    
    /**
     * Construct a new ZapiGenerator
     * @param packageName Root package for the generated classes
     * @throws IOException
     */
    public AutoZapiGenerator(String packageName, ZapidocParser parser, Properties properties) throws IOException {
        this(packageName, parser, new AutoZapiGeneratorFactory(packageName, properties));
        
    }
    
    public AutoZapiGenerator(String packageName, ZapidocParser parser, AutoZapiGeneratorFactory generatorFactory) throws IOException {
        this.packageName = packageName;
        this.parser = parser;
        this.generatorFactory = generatorFactory;
        
    }
    

    String getPackageName() {
        return this.packageName;
    }
    
    /**
     * Generate all class files into the supplied directory
     * @param outputDir Base classpath directory
     * @throws IOException 
     * @throws IOException
     */
    public void generateClassFiles(File outputDir) throws IOException {
        this.generateClassFiles(outputDir, null);
    }
    
    /**
     * Generate all class files in the supplied directory.
     * @param outputDir Base classpath directory
     * @param namespaces Set of namespace names to restrict output to. If a namespace zapi
     *    name differs from the java name (e.g. "com-e" vs. "come"), either name may be used.
     *    If this is null or empty, all known namespaces will be generated.
     * @throws IOException
     */
    public void generateClassFiles(File outputDir, Set<String> namespaces) throws IOException {
        this.generatorFactory.addWordReplacements(this.parser.getNamespaces());
        
        for (Namespace namespace : this.parser.getNamespaces()) {
            AutoZapiNamespaceGenerator namespaceGenerator = this.generatorFactory.getNamespaceGenerator(namespace);
            if ((namespaces == null) 
                    || namespaces.isEmpty()
                    || namespaces.contains(namespace.getName())
                    || namespaces.contains(namespaceGenerator.getNameGenerated())) {
                
                namespaceGenerator.generateClassFiles(outputDir);
            }
        }
        
        if (!this.parser.getErrdefs().isEmpty()) {
            AutoZapiErrdefGenerator generator = this.generatorFactory.getErrdefGenerator(this.parser.getErrdefs());
            File dir = new File(outputDir, ClassUtils.classNameToFilePath(this.packageName));
            generator.writeCode(dir);
        }
    }
    
    /**
     * Main entry point to create {@link com.netapp.common.zapi.ZAPIRunner}
     * annotated class files from zapidocs.<br>
     * <b>Usage:</b> ZapiGenerator &lt;zapidoc&gt; &lt;output-dir&gt; &lt;package&gt;<br>
     * <ul>
     *   <li>zapidoc: The zapidoc source file to parse
     *   <li>output-dir: The root classpath directory to output to
     *   <li>package: The root package to generate in
     * </ul>
     * For example, if "zapidoc.all" contains definitions for the "filer" namespace,
     * the command <samp>ZapiGenerator zapidoc.all src/generated com.netapp.zapi</samp> will create
     * classes in the "com.netapp.zapi.filer" package, in the "src/generated/com/netapp/zapi/filer/" directory.
     * The directory specified in output-dir must exist. Any directories required below that for packages 
     * will be created as needed.
     * @param args
     */
    public static void main(String[] args) {
        if (args.length < 4) {
            printUsage();
            System.exit(1);
        }

        String docFile = args[0];
        String sourceRoot = args[1];
        String packageName = args[2];
        
        // Only one property argument is expected here, so as to not conflict with the
        // variable number of 'namespace' arguments handled below. It would probably
        // make sense to have the 'namespace' arguments conform to the property
        // convention as well.
        List<String> propertyArgs = new ArrayList<String>();
        if (args.length > 3) {
            propertyArgs.add(args[3]);
        }

        Properties autoZapiGenerationProperties = GeneratorUtils.parsePropertyArgs(propertyArgs);

        Set<String> namespaces = null;
        if (args.length > 4) {
            namespaces = new HashSet<String>();
            for (int i = 4; i < args.length; i++) {
                namespaces.add(args[i]);
            }
        }
        
        try {
           // String classFileName = File.separator + ClassUtils.classNameToFilePath(packageName);
            File outputDir = new File(sourceRoot);

            if (!outputDir.exists() || !outputDir.isDirectory()) {
                throw new RuntimeException("Output source directory " + outputDir + " does not exist");            
            }


            ZapidocParser parser = new ZapidocParser(
                Arrays.asList(new File(docFile)));
            
            // TODO - add CLI processing and setup of Properties

            AutoZapiGenerator generator = new AutoZapiGenerator(packageName, parser, autoZapiGenerationProperties);

            generator.generateClassFiles(outputDir, namespaces);
        }
        catch (IOException e) {
            printUsage(e);
        }
    }
    
    /**
     * Print the usage of the main program
     *
     */
    static void printUsage() {
        System.err.println("Usage: AutoZapiGenerator <docs> <src_root> <package_root> <normal|vmo> <namespaces...>");
        System.err.println("    <normal|vmo> is whether SDK classes should be generated for normal Java Library use or to support VMO plugin use.");
    }
    
    /**
     * Print the usage of the main program
     * @param t
     */
    static void printUsage(Throwable t) {
        printUsage();
        if (t != null) {
            t.printStackTrace(System.err);
        }
        System.exit(1);
    }
}
